//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "FormMain.h"
#include "TrayIcon.h"
#include "FormAbout.h"
#include "FormSettings.h"
#include "FormAccount.h"
#include "FormHistory.h"
#include "FormSpeedDial.h"
#include "History.h"
#include "FormContacts.h"
#include "Contacts.h"
#include "FormTrayNotifier.h"
#include "LogUnit.h"
#include "Log.h"
#include "UaMain.h"
#include "ControlQueue.h"
#include "Callback.h"
#include "CallbackQueue.h"
#include <assert.h>
#include <DateUtils.hpp>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"

TfrmMain *frmMain;

namespace {
	History history;
	Contacts contacts;
	AnsiString asTransferHint = "Transfer to ...";
}

//---------------------------------------------------------------------------
__fastcall TfrmMain::TfrmMain(TComponent* Owner)
	: TForm(Owner)
{
	srand(time(0));
	lbl2ndParty->Caption = "";
	lbl2ndPartyDesc->Caption = "";
	lblCallState->Caption = "";
	frmHistory = new TfrmHistory(this->tsHistory, &history, &OnCall);
	frmHistory->Parent = tsHistory;
	frmHistory->Visible = true;
	frmSpeedDial = new TfrmSpeedDial(this->pnlSpeedDial, &OnCall);
	frmSpeedDial->Parent = this->pnlSpeedDial;
	frmSpeedDial->Visible = true;
	frmContacts = new TfrmContacts(this->tsContacts, &contacts, &OnCall);
	frmContacts->Parent = tsContacts;
	frmContacts->Visible = true;
	edTransfer->Text = asTransferHint;

	trIcon = new TrayIcon(this);
	trIcon->OnLeftBtnDown = OnTrayIconLeftBtnDown;
	trIcon->OnRightBtnDown = OnTrayIconRightBtnDown;
	trIcon->SetPopupMenu(popupTray);
	trIcon->SetIcon(Application->Icon);	
	trIcon->ShowInTray(true);
}

__fastcall TfrmMain::~TfrmMain()
{
	delete trIcon;
	trIcon = NULL;
}

//---------------------------------------------------------------------------
void __fastcall TfrmMain::FormCreate(TObject *Sender)
{
	AnsiString asHistoryFile;
	asHistoryFile.sprintf("%s\\%s_history.json", ExtractFileDir(Application->ExeName).c_str(),
		ChangeFileExt(ExtractFileName(Application->ExeName), "").c_str());
	history.SetFilename(asHistoryFile);
	history.Read();

	AnsiString asContactsFile;
	asContactsFile.sprintf("%s\\%s_contacts.json", ExtractFileDir(Application->ExeName).c_str(),
		ChangeFileExt(ExtractFileName(Application->ExeName), "").c_str());
	contacts.SetFilename(asContactsFile);
	contacts.Read();

#if 0
	if (this->BorderStyle != bsSingle)
	{
		this->Width = appSettings.frmMain.iWidth;
		this->Height = appSettings.frmMain.iHeight;
	}
#endif	
	this->Top = appSettings.frmMain.iPosY;
	this->Left = appSettings.frmMain.iPosX;
	if (appSettings.frmMain.bAlwaysOnTop)
		this->FormStyle = fsStayOnTop;
	else
		this->FormStyle = fsNormal;
	if (appSettings.frmMain.bWindowMaximized)
		this->WindowState = wsMaximized;
	if (appSettings.Logging.bLogToFile)
		CLog::Instance()->SetFile(ChangeFileExt(Application->ExeName, ".log").c_str());
	else
		CLog::Instance()->SetFile("");
	btnAutoAnswer->Down = appSettings.uaConf.autoAnswer;
	if (appSettings.frmMain.bSpeedDialVisible == false)
	{
		SetSpeedDial(appSettings.frmMain.bSpeedDialVisible);
	}
	UpdateCallHistory();
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::FormCloseQuery(TObject *Sender, bool &CanClose)
{
	AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".json" );
	appSettings.frmMain.bWindowMaximized = (this->WindowState == wsMaximized);
	if (!appSettings.frmMain.bWindowMaximized)
	{
		// these values are meaningless is wnd is maximized
		appSettings.frmMain.iWidth = this->Width;
		appSettings.frmMain.iHeight = this->Height;
		appSettings.frmMain.iPosY = this->Top;
		appSettings.frmMain.iPosX = this->Left;
	}
	appSettings.frmTrayNotifier.iPosX = frmTrayNotifier->Left;
	appSettings.frmTrayNotifier.iPosY = frmTrayNotifier->Top;
	appSettings.Write(asConfigFile);
	history.Write();

	CanClose = true;	
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::actShowAboutExecute(TObject *Sender)
{
	frmAbout->ShowModal();	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::actShowSettingsExecute(TObject *Sender)
{
    UaConf tmp = appSettings.uaConf;
	frmSettings->ShowModal();
	if (appSettings.uaConf != tmp)
	{
		SetStatus("Restarting UA...");
		miSettings->Enabled = false;		
		Ua::Instance().Restart();
	}
	if (appSettings.uaConf.logMessages != tmp.logMessages)
	{
		UA->SetMsgLogging(appSettings.uaConf.logMessages);
	}
	if (appSettings.frmMain.bAlwaysOnTop)
		this->FormStyle = fsStayOnTop;
	else
		this->FormStyle = fsNormal;
	if (appSettings.Logging.bLogToFile)
		CLog::Instance()->SetFile(ChangeFileExt(Application->ExeName, ".log").c_str());
	else
		CLog::Instance()->SetFile("");
	frmLog->SetLogLinesLimit(appSettings.Logging.iMaxUiLogLines);
	if (frmLog->Visible)
	{
    	frmLog->UpdateUi();
	}

	AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".json" );
	appSettings.Write(asConfigFile);
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::FormDestroy(TObject *Sender)
{
	UA->Destroy();
	UA_CB->Destroy();
	CLog::Instance()->Destroy();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::actShowLogExecute(TObject *Sender)
{
	if (!frmLog->Visible)
		frmLog->Show();
	frmLog->BringToFront();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::tmrStartupTimer(TObject *Sender)
{
	tmrStartup->Enabled = false;

	Ua::Instance().Start();
	SetStatus("Initializing...");	
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateCallHistory(void)
{
	cbCallURI->Clear();
	const std::deque<History::Entry>& entries = history.GetEntries();	
	for (unsigned int i=0; i<entries.size(); i++)
	{
		const History::Entry &entry = entries[i];
		if (entry.incoming == false)
		{
			// add only outgoing calls
			cbCallURI->Items->Add(entry.uri.c_str());
		}
	}
}

void __fastcall TfrmMain::btnMakeCallClick(TObject *Sender)
{
	if (call.incoming == false)
	{
		AnsiString target = cbCallURI->Text.Trim();
		if (target != "")
		{
			MakeCall(target);
		}
	}
	else
	{
    	UA->Answer(0);
	}
}
//---------------------------------------------------------------------------

void TfrmMain::MakeCall(AnsiString target)
{
	call.timestamp = Now();
	call.uri = target;
	UA->Call(0, target);
}

void __fastcall TfrmMain::btnHangupClick(TObject *Sender)
{
	Hangup();
}
//---------------------------------------------------------------------------

void TfrmMain::Hangup(void)
{
	UA->Hangup(0);
}

AnsiString TfrmMain::CleanUri(AnsiString uri)
{
	AnsiString res = uri;
	int start = uri.Pos("<");
	if (start == 0)
		start = 1;
	int end = uri.Pos(">");
	if (end == 0 || (uri.Pos(";") < end))
		end = uri.Pos(";");
	if (start && end && end > start)
	{
		res = uri.SubString(start, end-start);
	}
	return res;
}

void __fastcall TfrmMain::tmrCallbackPollTimer(TObject *Sender)
{
	Callback cb;
	if (UA_CB->GetCallback(cb) != 0)
	{
    	return;
	}
	switch (cb.type)
	{
		case Callback::CALL_STATE:
		{
			AnsiString asStateText;
			switch(cb.state)
			{
			case Callback::CALL_STATE_INCOMING:
				asStateText = "Incoming call";
				call.incoming = true;
				call.timestamp = Now();
				call.uri = cb.caller;
				if (appSettings.uaConf.autoAnswer)
				{
					int time = appSettings.uaConf.autoAnswerDelayMin;
					int delta = appSettings.uaConf.autoAnswerDelayMax - appSettings.uaConf.autoAnswerDelayMin + 1;
					unsigned int rand32 = (((unsigned int)rand()&0xFFFF)<<16) + (rand()&0xFFFF);
					if (delta)
					{
						time += rand32 % delta;
					}
					if (time == 0)
					{
						AutoAnswer();
						if (appSettings.uaConf.autoAnswerCode >= 400)
						{
							asStateText = "";
						}
					}
					else
					{
                        LOG("Delayed auto answer, time = %u ms\n", time);
						tmrAutoAnswer->Enabled = false;
						tmrAutoAnswer->Interval = time;
						tmrAutoAnswer->Enabled = true;
					}
				}
				break;
			case Callback::CALL_STATE_OUTGOING:
				asStateText = "Calling...";
				break;
			case Callback::CALL_STATE_TRYING:
				asStateText = "Trying...";
				break;
			case Callback::CALL_STATE_RINGING:
				asStateText = "Ringback";
				break;
			case Callback::CALL_STATE_PROGRESS:
				call.progress = true;
				asStateText = "Call state progress";
				break;
			case Callback::CALL_STATE_ESTABLISHED:
				asStateText = "Connected";
				call.connected = true;
				tmrAutoAnswer->Enabled = false;				
				break;
			case Callback::CALL_STATE_CLOSED: {
				asStateText = "";
				History::Entry entry;
				DecodeDateTime(call.timestamp,
					entry.timestamp.year, entry.timestamp.month, entry.timestamp.day,
					entry.timestamp.hour, entry.timestamp.min, entry.timestamp.sec,
					entry.timestamp.msec);
				entry.uri = call.uri.c_str();
				entry.incoming = call.incoming;
				history.AddEntry(entry);
				UpdateCallHistory();
				call.incoming = false;
				call.progress = false;
				call.connected = false;
				call.uri = "";
				btnHold->Down = false;
				tmrAutoAnswer->Enabled = false;				
				frmTrayNotifier->Close();
				break;
			}
			default:
				assert(!"Unhandled call state");
				break;
			}
			lbl2ndParty->Caption = CleanUri(cb.caller);
			lbl2ndPartyDesc->Caption = contacts.GetDescription(CleanUri(cb.caller));
			lblCallState->Caption = asStateText;

			if (cb.state == Callback::CALL_STATE_INCOMING)
			{
				frmTrayNotifier->SetText(lbl2ndPartyDesc->Caption, lbl2ndParty->Caption);
				if (appSettings.frmTrayNotifier.showOnIncoming)
				{
					if (Visible == false || appSettings.frmTrayNotifier.skipIfMainWindowVisible == false)
					{
					#if 1
						/** \todo frmTrayNotifier steals focus at first call */
						frmTrayNotifier->Visible = true;
					#else
						/** \todo frmTrayNotifier: no focus stealed, but inconsistent window state */
						ShowWindow(frmTrayNotifier->Handle,SW_SHOWNOACTIVATE);
						SetWindowPos(frmTrayNotifier->Handle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
					#endif
					}
				}
			}
			break;
		}
		case Callback::REG_STATE:
		{
			AnsiString asRegText;
			switch(cb.reg_state)
			{
			case Callback::REG_STATE_REGISTERING:
				asRegText = "Registering";
				break;
			case Callback::REG_STATE_REGISTER_OK:
				asRegText = "Registered";
				break;
			case Callback::REG_STATE_REGISTER_FAIL:
				asRegText = "Error registering";
				break;
			case Callback::REG_STATE_UNREGISTERING:
				asRegText = "Unregistering...";
				break;
			case Callback::REG_STATE_UNREGISTER_OK:
				asRegText = "Unregistered";
				break;
			case Callback::REG_STATE_UNREGISTER_FAIL:
				asRegText = "Unregistering error";
				break;
			default:
				assert(!"Unhandled reg state");
				break;
			}

			if (cb.reg_state != Callback::REG_STATE_UNREGISTERING &&
				cb.reg_state != Callback::REG_STATE_UNREGISTER_OK &&
				cb.reg_state != Callback::REG_STATE_UNREGISTER_FAIL)
				// skipping acc info for these states (current acc info may be different)
			{
				if (appSettings.uaConf.accounts.size())
				{
					asRegText.cat_sprintf(" %s@%s",
						appSettings.uaConf.accounts[0].user.c_str(),
						appSettings.uaConf.accounts[0].reg_server.c_str());
				}
			}
			SetStatus(asRegText);
			break;
		}
		case Callback::APP_STATE:
		{
			AnsiString text;
			/** \note Settings menu is disabled while initializing app */
			switch(cb.app_state)
			{
			case Callback::APP_INIT_FAILED:
				miSettings->Enabled = true;
				SetStatus("Failed to init application");
				break;
			case Callback::APP_INITIALIZED:
				miSettings->Enabled = true;
				text = "Initialized";
				if (appSettings.uaConf.accounts.size())
				{
					if (appSettings.uaConf.accounts[0].user.length())
					{
						text.cat_sprintf(": %s@%s",
							appSettings.uaConf.accounts[0].user.c_str(),
							appSettings.uaConf.accounts[0].reg_server.c_str());
					}
					else if (appSettings.uaConf.accounts[0].reg_server.length())
					{
						text.cat_sprintf(": %s",
							appSettings.uaConf.accounts[0].user.c_str(),
							appSettings.uaConf.accounts[0].reg_server.c_str());
					}
				}
				SetStatus(text);
				break;
			case Callback::APP_START_FAILED:
				SetStatus("Failed to start application");
				break;
			default:
				assert(!"Unhandled app state");
				break;
			}
			break;
		}
		case Callback::DLG_INFO_STATE:
		{
			frmSpeedDial->UpdateDlgInfoState(cb.contactId, cb.dlgInfoState);
			break;
		}
		default:
		{
			assert(!"Unhandled callback type");
		}
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnDialClick(TObject *Sender)
{
	char digit = '\0';
	TButton* btn = dynamic_cast<TButton*>(Sender);
	assert(btn);
	switch(btn->Tag)
	{
	case 0:
	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
	case 6:
	case 7:
	case 8:
	case 9:
		digit = '0' + btn->Tag;
		break;
	case 10:
		digit = '*';
		break;
	case 11:
		digit = '#';
		break;
	case 12:
		digit = 'A';
		break;
	case 13:
		digit = 'B';
		break;
	case 14:
		digit = 'C';
		break;
	case 15:
		digit = 'D';
		break;
	case 16:	// FLASH
		digit = 'R';
		break;
	default:
		assert(!"Unhandled button tag value");
	}

	if (call.connected || call.progress)
	{
    	UA->SendDigit(0, digit);
	}
	else
	{
		if (digit == 'R')
		{
			return;
		}
		cbCallURI->Text = cbCallURI->Text + digit;
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::FormClose(TObject *Sender, TCloseAction &Action)
{
    if(!Application->Terminated)
    {
        ShowWindow(Application->Handle, SW_HIDE);
        Visible = false;
        Action = caNone;
    }
	else
	{
		PostQuitMessage(0);
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnReregisterClick(TObject *Sender)
{
	UA->ReRegister(0);	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnAutoAnswerClick(TObject *Sender)
{
	appSettings.uaConf.autoAnswer = btnAutoAnswer->Down;	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnRedialClick(TObject *Sender)
{
	if (call.uri == "")
	{
		const std::deque<History::Entry>& entries = history.GetEntries();
		for (unsigned int i=0; i<entries.size(); i++)
		{
			const History::Entry &entry = entries[i];
			if (entry.incoming == false)
			{
				call.timestamp = Now();
				call.incoming = false;
				call.progress = false;
				call.uri = entry.uri.c_str();
				UA->Call(0, call.uri);
				break;
			}
		}
	}
}
//---------------------------------------------------------------------------

void TfrmMain::OnCall(AnsiString uri)
{
	if (call.uri != "")
	{
    	return;
	}
	call.timestamp = Now();
	call.incoming = false;
	call.progress = false;
	call.uri = uri.c_str();
	UA->Call(0, call.uri);
}

void __fastcall TfrmMain::edTransferEnter(TObject *Sender)
{
	if (edTransfer->Text == asTransferHint)
	{
		edTransfer->Text = "";
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnTransferClick(TObject *Sender)
{
	if (edTransfer->Text == asTransferHint || edTransfer->Text == "")
		return;
	UA->Transfer(0, edTransfer->Text);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnHoldClick(TObject *Sender)
{
	UA->Hold(0, btnHold->Down);	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnSpeedDialPanelClick(TObject *Sender)
{
	appSettings.frmMain.bSpeedDialVisible = !appSettings.frmMain.bSpeedDialVisible;
	SetSpeedDial(appSettings.frmMain.bSpeedDialVisible);
}
//---------------------------------------------------------------------------

void TfrmMain::SetSpeedDial(bool visible)
{
	if (visible)
	{
		this->Width += pnlSpeedDial->Width;	
		pnlSpeedDial->Visible = true;
		btnSpeedDialPanel->Caption = "<<";
	}
	else
	{
		pnlSpeedDial->Visible = false;
		this->Width -= pnlSpeedDial->Width;
		btnSpeedDialPanel->Caption = ">>";				
	}
}


void __fastcall TfrmMain::cbCallURIKeyPress(TObject *Sender, char &Key)
{
	if (Key == 13)
	{
		if (call.incoming == false)
		{
			AnsiString target = cbCallURI->Text.Trim();
			if (target != "")
			{
				MakeCall(target);
			}
		}
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnBackspaceClick(TObject *Sender)
{
	CallNumberBackspace();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnBackspaceMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
	tmrBackspace->Enabled = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnBackspaceMouseUp(TObject *Sender,
	  TMouseButton Button, TShiftState Shift, int X, int Y)
{
	tmrBackspace->Enabled = false;
}
//---------------------------------------------------------------------------

void TfrmMain::CallNumberBackspace(void)
{
	cbCallURI->Text = cbCallURI->Text.SubString(1, cbCallURI->Text.Length() - 1);
	cbCallURI->SelStart = cbCallURI->Text.Length();
}

void __fastcall TfrmMain::tmrBackspaceTimer(TObject *Sender)
{
	CallNumberBackspace();
}

//---------------------------------------------------------------------------
void __fastcall TfrmMain::OnTrayIconLeftBtnDown(TObject *Sender)
{
	Visible = !Visible;
	if (Visible)
	{
		Application->Restore();
		ShowWindow(Application->Handle, SW_SHOW);	// show taskbar button
		SetActiveWindow (Handle);
		SetForegroundWindow (Handle);
		SetWindowPos (Handle, HWND_TOP, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE);
	}
	else
	{
		ShowWindow(Application->Handle, SW_HIDE);	// hide taskbar button
	}
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::OnTrayIconRightBtnDown(TObject *Sender)
{
	//PopupMenu1 -> Tag = 1;
}

void __fastcall TfrmMain::miMinimizeTrayClick(TObject *Sender)
{
	Visible = false;
	ShowWindow(Application->Handle, SW_HIDE);	// hide taskbar button
}
//---------------------------------------------------------------------------

void TfrmMain::SetStatus(AnsiString text)
{
	StatusBar->SimpleText = text;
	trIcon->SetHint(text);
}

void __fastcall TfrmMain::actExitExecute(TObject *Sender)
{
	Ua::Instance().Quit();
	Application->Terminate();
	Close();
}
//---------------------------------------------------------------------------

void TfrmMain::AutoAnswer(void)
{
	if (appSettings.uaConf.autoAnswerCode == 200) {
		UA->Answer(0);
	} else if (appSettings.uaConf.autoAnswerCode >= 400) {
		UA->Hangup(0, appSettings.uaConf.autoAnswerCode);
		lbl2ndParty->Caption = "";
		lbl2ndPartyDesc->Caption = "";
		lblCallState->Caption = "";
		call.incoming = false;
		call.progress = false;
		call.connected = false;
		return;
	}
}

void __fastcall TfrmMain::tmrAutoAnswerTimer(TObject *Sender)
{
	AutoAnswer();
}
//---------------------------------------------------------------------------

